@@ -7,11 +7,12 @@ import aiosignal
 import pytest
 from multidict import CIMultiDict
 
-from aiohttp import WSMessage, WSMsgType
+from aiohttp import WSMessage, WSMessageTypeError, WSMsgType, web
+from aiohttp.http import WS_CLOSED_MESSAGE
 from aiohttp.streams import EofStream
-from aiohttp.test_utils import make_mocked_coro, make_mocked_request
+from aiohttp.test_utils import make_mocked_request
 from aiohttp.web import HTTPBadRequest, WebSocketResponse
-from aiohttp.web_ws import WS_CLOSED_MESSAGE, WebSocketReady
+from aiohttp.web_ws import WebSocketReady
 
 
 @pytest.fixture
@@ -67,6 +68,12 @@ async def test_nonstarted_pong() -> None:
         await ws.pong()
 
 
+async def test_nonstarted_send_frame() -> None:
+    ws = WebSocketResponse()
+    with pytest.raises(RuntimeError):
+        await ws.send_frame(b"string", WSMsgType.TEXT)
+
+
 async def test_nonstarted_send_str() -> None:
     ws = WebSocketResponse()
     with pytest.raises(RuntimeError):
@@ -243,6 +250,18 @@ def test_closed_after_ctor() -> None:
     assert ws.close_code is None
 
 
+async def test_raise_writer_limit(make_request) -> None:
+    """Test the writer limit can be adjusted."""
+    req = make_request("GET", "/")
+    ws = WebSocketResponse(writer_limit=1234567)
+    await ws.prepare(req)
+    assert ws._reader is not None
+    assert ws._writer is not None
+    assert ws._writer._limit == 1234567
+    ws._reader.feed_data(WS_CLOSED_MESSAGE, 0)
+    await ws.close()
+
+
 async def test_send_str_closed(make_request) -> None:
     req = make_request("GET", "/")
     ws = WebSocketResponse()
@@ -255,6 +274,21 @@ async def test_send_str_closed(make_request) -> None:
         await ws.send_str("string")
 
 
+async def test_recv_str_closed(make_request) -> None:
+    req = make_request("GET", "/")
+    ws = web.WebSocketResponse()
+    await ws.prepare(req)
+    assert ws._reader is not None
+    ws._reader.feed_data(WS_CLOSED_MESSAGE, 0)
+    await ws.close()
+
+    with pytest.raises(
+        WSMessageTypeError,
+        match=f"Received message {WSMsgType.CLOSED}:.+ is not WSMsgType.TEXT",
+    ):
+        await ws.receive_str()
+
+
 async def test_send_bytes_closed(make_request) -> None:
     req = make_request("GET", "/")
     ws = WebSocketResponse()
@@ -266,6 +300,21 @@ async def test_send_bytes_closed(make_request) -> None:
         await ws.send_bytes(b"bytes")
 
 
+async def test_recv_bytes_closed(make_request) -> None:
+    req = make_request("GET", "/")
+    ws = web.WebSocketResponse()
+    await ws.prepare(req)
+    assert ws._reader is not None
+    ws._reader.feed_data(WS_CLOSED_MESSAGE, 0)
+    await ws.close()
+
+    with pytest.raises(
+        WSMessageTypeError,
+        match=f"Received message {WSMsgType.CLOSED}:.+ is not WSMsgType.BINARY",
+    ):
+        await ws.receive_bytes()
+
+
 async def test_send_json_closed(make_request) -> None:
     req = make_request("GET", "/")
     ws = WebSocketResponse()
@@ -277,6 +326,18 @@ async def test_send_json_closed(make_request) -> None:
         await ws.send_json({"type": "json"})
 
 
+async def test_send_frame_closed(make_request) -> None:
+    req = make_request("GET", "/")
+    ws = WebSocketResponse()
+    await ws.prepare(req)
+    assert ws._reader is not None
+    ws._reader.feed_data(WS_CLOSED_MESSAGE, 0)
+    await ws.close()
+
+    with pytest.raises(ConnectionError):
+        await ws.send_frame(b'{"type": "json"}', WSMsgType.TEXT)
+
+
 async def test_ping_closed(make_request) -> None:
     req = make_request("GET", "/")
     ws = WebSocketResponse()
@@ -359,33 +420,112 @@ async def test_receive_eofstream_in_reader(make_request, loop) -> None:
 
     ws._reader = mock.Mock()
     exc = EofStream()
-    res = loop.create_future()
-    res.set_exception(exc)
-    ws._reader.read = make_mocked_coro(res)
-    ws._payload_writer.drain = mock.Mock()
-    ws._payload_writer.drain.return_value = loop.create_future()
-    ws._payload_writer.drain.return_value.set_result(True)
-
+    ws._reader.read = mock.AsyncMock(side_effect=exc)
+    assert ws._payload_writer is not None
+    f = loop.create_future()
+    f.set_result(True)
+    ws._payload_writer.drain.return_value = f  # type: ignore[attr-defined]
     msg = await ws.receive()
     assert msg.type == WSMsgType.CLOSED
     assert ws.closed
 
 
-async def test_receive_timeouterror(make_request, loop) -> None:
+async def test_receive_exception_in_reader(make_request: Any, loop: Any) -> None:
+    req = make_request("GET", "/")
+    ws = WebSocketResponse()
+    await ws.prepare(req)
+
+    ws._reader = mock.Mock()
+    exc = Exception()
+    ws._reader.read = mock.AsyncMock(side_effect=exc)
+
+    msg = await ws.receive()
+    assert msg.type == WSMsgType.ERROR
+    assert ws.closed
+    assert len(ws._req.transport.close.mock_calls) == 1
+
+
+async def test_receive_close_but_left_open(make_request: Any, loop: Any) -> None:
+    req = make_request("GET", "/")
+    ws = WebSocketResponse()
+    await ws.prepare(req)
+    close_message = WSMessage(WSMsgType.CLOSE, 1000, "close")
+
+    ws._reader = mock.Mock()
+    ws._reader.read = mock.AsyncMock(return_value=close_message)
+    ws._payload_writer.drain = mock.Mock()
+    ws._payload_writer.drain.return_value = loop.create_future()
+    ws._payload_writer.drain.return_value.set_result(True)
+
+    msg = await ws.receive()
+    assert msg.type == WSMsgType.CLOSE
+    assert ws.closed
+    assert len(ws._req.transport.close.mock_calls) == 1
+
+
+async def test_receive_closing(make_request: Any, loop: Any) -> None:
+    req = make_request("GET", "/")
+    ws = WebSocketResponse()
+    await ws.prepare(req)
+    closing_message = WSMessage(WSMsgType.CLOSING, 1000, "closing")
+
+    ws._reader = mock.Mock()
+    read_mock = mock.AsyncMock(return_value=closing_message)
+    ws._reader.read = read_mock
+    ws._payload_writer.drain = mock.Mock()
+    ws._payload_writer.drain.return_value = loop.create_future()
+    ws._payload_writer.drain.return_value.set_result(True)
+
+    msg = await ws.receive()
+    assert msg.type == WSMsgType.CLOSING
+    assert not ws.closed
+
+    msg = await ws.receive()
+    assert msg.type == WSMsgType.CLOSING
+    assert not ws.closed
+
+    ws._cancel(ConnectionResetError("Connection lost"))
+
+    msg = await ws.receive()
+    assert msg.type == WSMsgType.CLOSING
+
+
+async def test_close_after_closing(make_request: Any, loop: Any) -> None:
+    req = make_request("GET", "/")
+    ws = WebSocketResponse()
+    await ws.prepare(req)
+    closing_message = WSMessage(WSMsgType.CLOSING, 1000, "closing")
+
+    ws._reader = mock.Mock()
+    ws._reader.read = mock.AsyncMock(return_value=closing_message)
+    ws._payload_writer.drain = mock.Mock()
+    ws._payload_writer.drain.return_value = loop.create_future()
+    ws._payload_writer.drain.return_value.set_result(True)
+
+    msg = await ws.receive()
+    assert msg.type == WSMsgType.CLOSING
+    assert not ws.closed
+    assert len(ws._req.transport.close.mock_calls) == 0
+
+    await ws.close()
+    assert ws.closed
+    assert len(ws._req.transport.close.mock_calls) == 1
+
+
+async def test_receive_timeouterror(make_request: Any, loop: Any) -> None:
     req = make_request("GET", "/")
     ws = WebSocketResponse()
     await ws.prepare(req)
     assert len(ws._req.transport.close.mock_calls) == 0
 
     ws._reader = mock.Mock()
-    res = loop.create_future()
-    res.set_exception(asyncio.TimeoutError())
-    ws._reader.read = make_mocked_coro(res)
+    ws._reader.read = mock.AsyncMock(side_effect=asyncio.TimeoutError())
 
     with pytest.raises(asyncio.TimeoutError):
         await ws.receive()
 
-    assert len(ws._req.transport.close.mock_calls) == 1
+    # Should not close the connection on timeout
+    assert len(ws._req.transport.close.mock_calls) == 0
 
 
 async def test_multiple_receive_on_close_connection(make_request) -> None:
@@ -448,16 +588,18 @@ async def test_send_with_per_message_deflate(make_request, mocker) -> None:
     req = make_request("GET", "/")
     ws = WebSocketResponse()
     await ws.prepare(req)
-    writer_send = ws._writer.send = make_mocked_coro()
+    with mock.patch.object(ws._writer, "send_frame", autospec=True, spec_set=True) as m:
+        await ws.send_str("string", compress=15)
+        m.assert_called_with(b"string", WSMsgType.TEXT, compress=15)
 
-    await ws.send_str("string", compress=15)
-    writer_send.assert_called_with("string", binary=False, compress=15)
+        await ws.send_bytes(b"bytes", compress=0)
+        m.assert_called_with(b"bytes", WSMsgType.BINARY, compress=0)
 
-    await ws.send_bytes(b"bytes", compress=0)
-    writer_send.assert_called_with(b"bytes", binary=True, compress=0)
+        await ws.send_json("[{}]", compress=9)
+        m.assert_called_with(b'"[{}]"', WSMsgType.TEXT, compress=9)
 
-    await ws.send_json("[{}]", compress=9)
-    writer_send.assert_called_with('"[{}]"', binary=False, compress=9)
+        await ws.send_frame(b"[{}]", WSMsgType.TEXT, compress=9)
+        m.assert_called_with(b"[{}]", WSMsgType.TEXT, compress=9)
 
 
 async def test_no_transfer_encoding_header(make_request, mocker) -> None:
@@ -497,4 +639,4 @@ async def test_get_extra_info(
     await ws.prepare(req)
     ws._writer = ws_transport
 
-    assert ws.get_extra_info(valid_key, default_value) == expected_result
+    assert expected_result == ws.get_extra_info(valid_key, default_value)
